;EasyCodeName=modFPU,1
;*****************************************************************************
;				DSP and Fast Fourier Transforms - Intel FPU unit
;*****************************************************************************
;Acknowledgements:
;	Dr. Steve Smith			- 	Digital Signal Processing
;	Raymond Filiatreault 	- 	Simply FPU
;	Hutch					-	The Masm Website
;	The other people on the Masm website, for their free info and examples
;*****************************************************************************
;*****************************************************************************
;							THE WAY TO DO IT
;							----------------
;	All Signal and FFT arrays are initialised external to this module. You
;	provide the array pointers and the array sizes.
;
;1)	CreateBRTable:-			Allocates memory for table, Populates it and Bit
;							Reverses the table pointers.
;							You only have to call proc once on initialisation
;							only.
;
;2)	Convert_Int16_Real8:-	If necessary Convert 16 bit integers to 64bit
;							Real8 FP values.
;
;3)	Convert_Int32_Real8:-	If necessary Convert 32 bit integers to 64bit
;							Real8 FP values.
;
;4)	BitReverse_Array:-		Before you apply the FFT to an array, you must
;							BitReverse it. This is external to the FFT for
;							speed optimisation.
;
;5) Real8_FFT:-				The FFT, based on Steve Smith's example in his
;							DSP book. Operates on 64bit Real8 FP values.
;
;*****************************************************************************

.Const
;-----------------------------------------------------------------------------
;Bit reversal table array sizes
;-----------------------------------------------------------------------------
BitRevSize_32		Equ		32			;2^N
BitRevSize_64		Equ		64			;
BitRevSize_128		Equ		128			;
BitRevSize_256		Equ		256			;
BitRevSize_512		Equ		512			;
BitRevSize_1024		Equ		1024		;
BitRevSize_2048		Equ		2048		;
BitRevSize_4096		Equ		4096		;
BitRevSize_8192		Equ		8192		;
BitRevSize_16384	Equ		16384		;
BitRevSize_32678	Equ		32768 		;
BitRevSize_65536	Equ		65536		;

BitRevTablesize		Equ		523800		;130950 * 4	bytes

BitRevSwop1			Equ		55555555H	;Bit Reverse Shift masks
BitRevSwop1Not		Equ		0AAAAAAAAH	;
BitRevSwop2			Equ		33333333H	;
BitRevSwop2Not		Equ		0CCCCCCCCH	;
BitRevSwop4			Equ		0F0F0F0FH	;
BitRevSwop4Not		Equ		0F0F0F0F0H	;
BitRevSwop8			Equ		00FF00FFH	;
BitRevSwop8Not		Equ		0FF00FF00H	;
BitRevSwop16		Equ		0000FFFFH	;
BitRevSwop16Not		Equ		0FFFF0000H	;

;-----------------------------------------------------------------------------
;FPU Control Register Bits - Extracted from 'Simply FPU' and Intel
;-----------------------------------------------------------------------------
CR_InfinityControl	Equ		1000H		;This field has been retained For
										;compatibility with the 287 and
										;earlier co - processors.In the more
										;modern FPUs, This bit is disregarded
										;and both - infinity and + infinity
										;are respected.

;0 = Both - infinity and + infinity are treated as unsigned infinity (Default)
;1 = Respects both -infinity and +infinity

CR_RoundingControl	Equ		0C00H		;The RC field (bits 11 and 10) or
										;Rounding Control determines how the
										;FPU will round results in one of
										;four ways

;00 = Round to nearest, or to even if equidistant (Default)
;01 = Round down (toward -infinity)
;10 = Round up (toward +infinity)
;11 = Truncate (toward 0)

CR_PrecisionControl	Equ		0300H		;The PC field (bits 9 and 8) or
										;Precision Control determines to what
										;precision the FPU rounds results
										;after each arithmetic instruction in
										;one of three ways
CR_PrecisionAND		Equ		00CFH		;
CR_PrecisionReal4	Equ		0000H		;
CR_PrecisionReal8	Equ		0020H		;
CR_PrecisionReal10	Equ		0030H		;


;00 = 24 bits (Real4)
;01 = Not used
;10 = 53 bits (REAL8)
;11 = 64 bits (REAL10) (Default)

CR_IntEnableMask	Equ		0080H		;The IEM field (bit 7) or Interrupt
										;Enable Mask determines whether any
										;of the interrupt masks will be
										;enabled.
										;This field is also For compatibility
										;with early co - processors and not
										;used anymore.

;0 = Interrupts enabled
;1 = Interrupts disabled (Default)

;INTERRUPT MASK BITS
CR_PrecisionMask	Equ		0020H		;
CR_UnderflowMask	Equ		0010H		;
CR_OverflowMask		Equ		0008H		;
CR_ZeroDivideMask	Equ		0004H		;
CR_DenormaliseMask	Equ		0002H		;
CR_InvalidOpMask	Equ		0001H		;

;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;FPU Status Register Bits - Extracted from 'Simply FPU' and Intel
;-----------------------------------------------------------------------------
;When the FPU is intialised all bits set to 0

ST_FPUBusy			Equ		8000H		;FPU is busy executing

ST_ConditionCodes	Equ		4700H		;Instructions change these

ST_TopOfstack		Equ		3800H		;FPU barrel top of stack

ST_InterruptRequest	Equ		0080H		;set to 1 by the FPU while an
										;exception is being handled and gets
;reset to 0 when the exceptionhandli ng is completed.
;When the interrupt is masked in the Control Word For the FPU to handle the
;exception, This bit may never be seen to be set While stepping through the
;instructions with a debugger. However, IF the programmer handles the
;interrupt, that bit should remain set until the interrupt handling routine is
;completed.

;Bits 6 - 0 are flags raised by the FPU whenever it detects an exception.
;Those exception flags are cumulative in the sense that, once set (bit = 1),
;they are not reset (bit = 0) by the result of a subsequent instruction which,
;by itself, would not have raised that flag. Those flags can only be reset by
;either initializing the FPU (FINIT instruction) or by explicitly clearing
;those flags (FCLEX instruction).

ST_StackFault		Equ		0040H		;set whenever an attempt is made to
										;either load a value into a register
;which is not free (the C1 bit would also get set to 1) or pop a value from
;a register which is free (and the C1 bit would get reset to 0).
;(Such stack fault is also treated as an invalid operation and the I field
;flag bit0 would thus also be set by this exception; see below.)

ST_PrecisionExcept	Equ		0020H		;set whenever some precision is lost
										;by instructions which do exact
;arithmetic. For example, dividing 1 by 10 does not yield an exact value in
;binary arithmetic and would set the P exception flag. Another example which
;sets the P exception flag would be the conversion of a REAL10 to a REAL4
;when some of the least significant bits would be lost. If the FPU handles
;this exception (when the PM bit is set in the Control Word), it rounds the
;result according to the rounding mode specified in the RC field of the
;Control Word.

ST_Underflow		Equ		0010H		;set whenever a value is too small
			 							;(without being equal to 0) to be
;represented properly. Each of the floating point formats has a different
;limit on the smallest number which can be represented. The U flag gets
;set IF the result of an operation exceeds that limit. For example,
;dividing a valid very small number by a large number could exceed the
;limit. A valid Real10 small number may be much smaller than acceptable.
;For the Real4 or Real8 formats ; in such cases, conversion from the former
;to the latter would also set the U flag. If the FPU handles this exception
;(when the UM bit is set in the Control Word), it would denormalize the value
;until the exponent is in range or ultimately return a 0.


ST_Overflow			Equ		0008H		;set whenever a value is too large in
								 		;magnitude to be represented properly.
;Again, each of the floating point formats has a different limit on the
;largest number which can be represented. The O flag gets set IF the result of
;an operation exceeds that limit. For example, multiplying a valid very large
;number by another large number could exceed the limit. A valid Real10 large
;number may be much larger than acceptable For the Real4 or Real8 formats
;conversion from the former to the latter would also set the O flag.
;If the FPU handles this exception(when the OM bit is set in the Control Word)
;it would generate a properly signed INFINITY according to the IC flag of the
;Control Word.

ST_ZeroDivide		Equ		0004H		;set whenever the division of a finite
										;non-zero value by 0 is attempted.
;If the FPU handles this exception (when the ZM bit is set in the Control Word)
;it would generate a properly signed INFINITY according to the XOR of the
;operand signs and then according to the IC flag of the Control Word.


ST_Denormalised		Equ		0002H		;set whenever an instruction attempts
										;to operate on a denormalized number
;or the result of the operation is a denormalized number.
;If the FPU handles this exception (when the DM bit is set in the Control Word)
;it would simply continue with normal processing and then check for other
;possible exceptions.

ST_InvalidOp		Equ		0001H		;set whenever an operation is
										;considered invalid by the FPU.
; Examples of such operations are:
;- Stack overflow or underflow
;- Indeterminate arithmetic such as 0 divided by 0, or subtracting infinity
;  from infinity
;- Using a Not-A-Number (NAN) as an operand with some instructions
;- Trying to extract the square root of a negative number

;If the FPU handles this exception (when the IM bit is set in the Control Word)
;it would either return the NAN if one was used as an operand (or the larger
;absolute value if two NANs were used as operands), or otherwise return a
;special "INDEFINITE" NAN.

;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;FPU Tag Register Bits - Extracted from 'Simply FPU' and Intel
;-----------------------------------------------------------------------------

Tag_7				Equ		0C000H		;
Tag_6				Equ		3000H		;
Tag_5				Equ		0C00H		;
Tag_4				Equ		0300H		;
Tag_3				Equ		00C0H		;
Tag_2				Equ		0030H		;
Tag_1				Equ		000CH		;
Tag_0				Equ		0003H		;

;00 = The register contains a valid non - zero value
;01 = The register contains a value equal to 0
;10 = The register contains a special value (NAN, infinity, or denormal)
;11 = The register is empty

;When the FPU is initialized, All the 80 - bit registers are empty and the Tag
;Word would thus have an overall value of 1111111111111111b (FFFFh).

;If a valid non-zero value is then loaded, the Tag Word would then be change to
;0011111111111111b (3FFFh). (Remember that the very first value loaded goes
;into BC7.)

;If a second value equal to 0 was then loaded, the Tag Word would become
;0001111111111111b (1FFFh). (And the second value loaded goes into BC6.)

;Although this Tag Word may contain information which could also be useful to
;the programmer, it cannot be accessed directly nor by itself. The only way to
;gain access to it is to store the FPU's environment data in memory
;(see the FSTENV instruction) and examine it there. However, the information
;available in the Tag Word could also be obtained otherwise
;(such as with the FXAM instruction for individual registers).

.Data?
;-----------------------------------------------------------------------------
;BIT REVERSAL TABLE POINTERS
;-----------------------------------------------------------------------------
BitRevTableHwin		DD	?	;ALLOC function handle return
BitRevTablePtr		DD	?	;Base pointer from ALLOC function

BitRevPtr_32		DD	?	;Pointers to individual bitreversal table offsets
BitRevPtr_64		DD	?	;
BitRevPtr_128		DD	?	;
BitRevPtr_256		DD	?	;
BitRevPtr_512		DD	?	;
BitRevPtr_1024		DD	?	;
BitRevPtr_2048		DD	?	;
BitRevPtr_4096		DD	?	;
BitRevPtr_8192		DD	?	;
BitRevPtr_16384		DD	?	;
BitRevptr_32678		DD	?	;
BitRevptr_65536		DD	?	;


;-----------------------------------------------------------------------------
;FPU state save structure 
;Instructions used:- FSAVE / FNSAVE
;-----------------------------------------------------------------------------

FPUState Struct
	ControlWord			DD	?	;16 bit control word in lower WORD
	StatusWord			DD	?	;16 bit status word in lower WORD
	TagWord				DD	?	;16 bit tag word in lower WORD
	InstuctionPtr		DD	?	;32 bit instruction pointer
	CodeSegment			DD	?	;16 bit code segment in lower WORD
	OperandAddr			DD	?	;32 bit operand address
	DataSegment			DD	?	;16 bit data segment in lower WORD
	FPU_ST0				DT	?	;80 bit barrel register 0
	FPU_ST1				DT	?	;80 bit barrel register 1
	FPU_ST2				DT	?	;80 bit barrel register 2
	FPU_ST3				DT	?	;80 bit barrel register 3
	FPU_ST4				DT	?	;80 bit barrel register 4
	FPU_ST5				DT	?	;80 bit barrel register 5
	FPU_ST6				DT	?	;80 bit barrel register 6
	FPU_ST7				DT	?	;80 bit barrel register 7
FPUState EndS


FPUControlReg			DD	?	;FPU control temp word
FFT_Memory1				DD	?	;Temporary memory
FFT_Inputsize			DD	?	;FFT array size
FFT_InputsizeD2			DD	?	;FFT input size divide 2

FFT_StageLoopLimit		DD	?	;FFT alg variables
FFT_StageCounter		DD	?	;
FFT_DFTLoopLimit		DD	?	;
FFT_DFTLoopCounter		DD	?	;
FFT_ButterFlyStep		DD	?	;
FFT_ButterFlyCounter	DD	?	;
FFT_ButterFlyLoopLimit	DD	?	;
FFT_IPtr				DD	?	;

FFT_URDbl				DT	?	;
FFT_UIDbl				DT	?	;
FFT_TempDbl				DT	?	;
FFT_CosTmp				DT	?	;
FFT_SinTmp				DT	?	;

.Data

.Code

;-----------------------------------------------------------------------------
;CreateBRTable:		Creates a Bit Reversal Table which is used for BR'ing any
;			   		FFT aray.
;
;			   		Consists of arrays of DW offset pointers for tablesizes of
;			   		32,64,128,256,....65536.
;
;Called Procs:		BR_Populate - Sets original pointers
;					Do_BitReversal - BR's the whole table
;
;I/P Params:		None
;O/P Params:		BitRevTablePtr - Contains BR table offset
;					BitRevPtr_32(etc) - Individual arraysize offsets into table
;
;Return Values:		True - All is OK
;					False - Could not allocate memory for table
;-----------------------------------------------------------------------------
CreateBRTable Proc

	Push Ebx									;Using EBX
;----------- FIND SOME MEMORY SPACE ------------
	Invoke LocalAlloc, LPTR, BitRevTablesize	;523800	= 130950 * 4	bytes
	Jnz @F										;NULL = error
	Ret 										;Return FALSE

;---------- CALCULATE TABLE POINTERS -----------
@@:												;
	Mov BitRevTablePtr, Eax						;Base pointer
	Mov BitRevPtr_32, Eax						;Table 32

	Mov Ebx, 32									;ebx = 32 ?????
	Shl Ebx, 2									;ebx = 32 x 4 - DWORD sizes
	Add Eax, Ebx 								;
	Mov BitRevPtr_64, Eax						;Table 64 (80H)

	Shl Ebx, 1									;ebx = 64 x 4
	Add Eax, Ebx 								;
	Mov BitRevPtr_128, Eax						;Table 128 (180H)

	Shl Ebx, 1									;ebx = 128 x 4
	Add Eax, Ebx 								;
	Mov BitRevPtr_256, Eax						;Table 256 (380H)

	Shl Ebx, 1									;ebx = 256 x 4
	Add Eax, Ebx 								;
	Mov BitRevPtr_512, Eax						;Table 512 (780H)
  
	Shl Ebx, 1									;ebx = 512 x 4
	Add Eax, Ebx 								;
	Mov BitRevPtr_1024, Eax						;Table 1024 (F80H)

	Shl Ebx, 1									;ebx = 1024 x 4
	Add Eax, Ebx 								;
	Mov BitRevPtr_2048, Eax						;Table 2048

	Shl Ebx, 1									;ebx = 2048 x 4
	Add Eax, Ebx 								;
	Mov BitRevPtr_4096, Eax						;Table 4096

	Shl Ebx, 1									;ebx = 4096 x 4
	Add Eax, Ebx 								;
	Mov BitRevPtr_8192, Eax						;Table 8192

	Shl Ebx, 1									;ebx = 8192 x 4
	Add Eax, Ebx 								;
	Mov BitRevPtr_16384, Eax					;Table 16384

	Shl Ebx, 1									;ebx = 16384 x 4
	Add Eax, Ebx 								;
	Mov BitRevptr_32678, Eax					;Table 32678

	Shl Ebx, 1									;ebx = 32678 x 4
	Add Eax, Ebx 								;
	Mov BitRevptr_65536, Eax					;Table 65536

;--------POPULATE TABLES WITH OFFSETS-----------

	Invoke	BR_Populate, BitRevSize_32, BitRevPtr_32
	Invoke	BR_Populate, BitRevSize_64, BitRevPtr_64
	Invoke	BR_Populate, BitRevSize_128, BitRevPtr_128
	Invoke	BR_Populate, BitRevSize_256, BitRevPtr_256
	Invoke	BR_Populate, BitRevSize_512, BitRevPtr_512
	Invoke	BR_Populate, BitRevSize_1024, BitRevPtr_1024
	Invoke	BR_Populate, BitRevSize_2048, BitRevPtr_2048
	Invoke	BR_Populate, BitRevSize_4096, BitRevPtr_4096
	Invoke	BR_Populate, BitRevSize_8192, BitRevPtr_8192
	Invoke	BR_Populate, BitRevSize_16384, BitRevPtr_16384
	Invoke	BR_Populate, BitRevSize_32678, BitRevptr_32678
	Invoke	BR_Populate, BitRevSize_65536, BitRevptr_65536

;--------THE ACTUAL BIT REVERSAL PART-----------

	Invoke	Do_BitReversal						;Push bits around

	Pop Ebx
	Mov Eax, TRUE								;
	Ret 										;
CreateBRTable EndP

;-----------------------------------------------------------------------------
;BR_Populate:		Place offset pointers in the BR table
;
;I/P Params:		mySize - This table size
;					myPtr - This table pointer
;
;O/P Params:		True - Populated table
;
;Return Values:		True - All is OK
;-----------------------------------------------------------------------------
BR_Populate Proc mySize:DWord, myPtr:DWord


	Push Edi									;Prologue
	Push Ecx									;
;-----------------------------------------------
	Mov Edi, myPtr								;Load up
	Mov Ecx, mySize								;
	Dec Ecx										;Zero base adjust and correct
												;ptr value
BRP_Loop:
	Mov [Edi + Ecx * 4], Ecx					;
	Loop BRP_Loop								;And around again
	Mov [Edi], 0								;Last location
;-----------------------------------------------
	Pop Ecx										;Epilogue
	Pop Edi										;

	Mov Eax, TRUE								;
	Ret 										;
BR_Populate EndP

;-----------------------------------------------------------------------------
;Do_BitReversal:	Bit Reverses the whole BR table.
;
;I/P Params:		Private variables (BitRevTablePtr, BitRevTablesize) must
;					valid values.
;
;O/P Params:		True - Table has been reversed for REAL8.
;
;Return Values:		True - All is OK
;-----------------------------------------------------------------------------
Do_BitReversal Proc

	Push Esi									;Prologue
	Push Ecx									;
	Push Edx									;
;-----------------------------------------------
	Mov Esi, BitRevTablePtr						;Base of Bit Reverse table
	Sub Esi, 4									;Adjust for zero base
	Mov Ecx, BitRevTablesize					;As it says

BR_Loop:
	Mov Eax, [Esi + Ecx * 4]					;Get value
	And Eax, 0000FFFFH							;CURRENTLY ONLY 16BIT POINTERS
	Mov Edx, Eax								;EDX = EAX

;--- Level 1 Swop ---							;1 bit swop
	And Edx, BitRevSwop1						;Right hand bits
	Shl Edx, 1									;Shift left one bit
	And Eax, BitRevSwop1Not						;Left hand bits
	Shr Eax, 1									;Shift right one bit
	Or Eax, Edx									;EAX = swopped bits
	Mov Edx, Eax								;EDX = EAX

;--- Level 2 Swop ---							;2 bit swop
	And Edx, BitRevSwop2						;
	Shl Edx, 2									;
	And Eax, BitRevSwop2Not						;
	Shr Eax, 2									;
	Or Eax, Edx									;
	Mov Edx, Eax								;

;--- Level 4 Swop ---							;4 bit swop
	And Edx, BitRevSwop4						;
	Shl Edx, 4									;
	And Eax, BitRevSwop4Not						;
	Shr Eax, 4									;
	Or Eax, Edx									;
	Mov Edx, Eax								;

;--- Level 8 Swop ---							;8 bit swop
	And Edx, BitRevSwop8						;
	Shl Edx, 8									;
	And Eax, BitRevSwop8Not						;
	Shr Eax, 8									;
	Or Eax, Edx									;
	Mov Edx, Eax								;Now it looks a mess, so..

	Mov [Esi + Ecx * 4], Eax					;Put it back where you found
												;it so nobody will notice.
	Loop BR_Loop								;what!.. again ?
;-----------------------------------------------
	Pop Edx										;Epilogue
	Pop Ecx										;
	Pop Esi										;

	Mov Eax, TRUE								;
	Ret 										;
Do_BitReversal EndP

;-----------------------------------------------------------------------------
;BitReverse_Signal:		Bit Reverses the supplied signal of REAL8s
;
;I/P Params:			InputArray_Ptr - Source signal
;						OutputArray_Ptr - Destination array
;						InputSize - Size of array
;
;O/P Params:			OutputArray Bit Reversed
;						True - OK
;						False - Wrong Arraysize (must be 32 <= 2^N <= 65536)
;-----------------------------------------------------------------------------
BitReverse_Array Proc InputSize:DWord, InputArray_Ptr:DWord, OutputArray_Ptr:DWord

	Push Ebx									;Prologue
	Push Ecx									;
	Push Edx									;
	Push Esi									;
	Push Edi									;

;-----LOAD SOURCE AND DESTINATION POINTERS------
	Mov Esi, InputArray_Ptr						;Signal array source
	Sub Esi, 8
	Mov Edi, OutputArray_Ptr					;
	Mov Ecx, InputSize							;

;--CHECK SIZE OF SIGNAL AND APPLY TABLE OFFSETS-
	Cmp Ecx, BitRevSize_32						;Array size = 32 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_32						;
	Jmp RIS_Loop								;
@@:
	Cmp Ecx, BitRevSize_64						;Array size = 64 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_64						;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_128						;Array size = 128 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_128						;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_256						;Array size = 256 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_256						;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_512						;Array size = 512 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_512						;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_1024					;Array size = 1024 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_1024						;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_2048					;Array size = 2048 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_2048						;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_4096					;Array size = 4096 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_4096						;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_8192					;Array size = 8192 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_8192						;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_16384					;Array size = 16384 ?
	Jne @F										;
	Mov Ebx, BitRevPtr_16384					;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_32678					;Array size = 32678 ?
	Jne @F										;
	Mov Ebx, BitRevptr_32678					;
	Jmp RIS_Loop								;
@@:
   	Cmp Ecx, BitRevSize_65536					;Array size = 65536 ?
	Je @F										;
;-----------------------------------------------
	Pop Edi										;Epilogue
	Pop Esi										;
	Pop Edx										;
	Pop Ecx										;
	Pop Ebx										;
	Mov Eax, FALSE								;Fail - Wrong array size
	Ret 										;
;-----------------------------------------------

@@:
	Mov Ebx, BitRevptr_65536					;
;-----------------------------------------------
;BR Table offsets are 32 bit.
;
;ESI - Input signal (Real8) pointer
;EDI - Output signal (Real8) pointer
;EBX - BR table offset/Output signal offset
;ECX - Signal size
;-----------------------------------------------
	Sub Ebx, 4									;That persistent zero base adj

RIS_Loop:
	Mov Eax, [Ebx + Ecx * 4]					;EAX = BR offset

	Mov Edx, [Esi + Eax * 8]					;EDX = InputArray lo-Dword
	Mov [Edi + Eax * 8], Edx					;OutPutArray = BReversed EAX

	Mov Edx, [Esi][Eax * 8 + 4]					;EDX = InputArray hi-Dword
	Mov [Edi][Eax * 8 + 4], Edx					;OutPutArray = BReversed EAX

	Loop RIS_Loop								;Done??

;--- Maybe try this method later  --------------
;RIS_Loop:
;	Mov Eax, [Ebx + Ecx * 4]					;EAX = BR offset
;	Fld QWord Ptr [Esi + Ecx * 8]				;(0)=Value
;	Fstp QWord Ptr [Edi + Eax * 8]				;Barrel clear!
;	Loop RIS_Loop								;Done??
;-----------------------------------------------
	Pop Edi										;Epilogue
	Pop Esi										;
	Pop Edx										;
	Pop Ecx										;
	Pop Ebx										;


	Mov Eax, TRUE								;
	Ret 										;
BitReverse_Array EndP
;-----------------------------------------------------------------------------
;Convert_Int16_Real8:		Converts an array of 16bit integers (input array)
;					 		into Real8 FP values (output array)
;
;-----------------------------------------------------------------------------
Convert_Int16_Real8 Proc InputSize:DWord, InputArray_Ptr:DWord, OutputArray:DWord


;---------SET UP FPU FOR OPERATION----------
	Fsave FPUState							;Just save the whole bang shoot
	Fwait									;zzz...
	Finit									;BANG!!
	Fwait

	Fstcw FPUControlReg						;Load Control word
	Fwait
	Mov Eax, CR_PrecisionAND				;Select Real8 precision mask
	And Eax, FPUControlReg					;Clr precision bits
	Or Eax, CR_PrecisionReal8				;Set Real8 precision
	Mov FPUControlReg, Eax					;Put in memory location
	Fldcw FPUControlReg						;Real8 mode
	Fwait

;----LOAD SOURCE AND DESTINATION POINTERS---
	Mov Esi, InputArray_Ptr					;Signal array source
	Sub Esi, 2								;Zero base adjust
	Mov Edi, OutputArray_Ptr				;
	Sub Edi, 8								;Zero base adjust
	Mov Ecx, InputSize						;

Con16Loop:
	Fild Word Ptr [Esi + Ecx * 2]			;Load 16bit integer
	Fwait									;zzz...
	Fstp QWord Ptr [Edi + Ecx * 8]			;Save as 64bit Real8
	Fwait									;
	Loop Con16Loop							;

;------------RESTORE FPU STATE--------------
	Frstor FPUState							;Restore FPU state
	Fwait									;

	Mov Eax, TRUE							;
	Ret 									;
Convert_Int16_Real8 EndP
;-----------------------------------------------------------------------------
;Convert_Int32_Real8:		Converts an array of 32bit integers (input array)
;					 		into 64bit FP values (output array)
;
;-----------------------------------------------------------------------------
Convert_Int32_Real8 Proc InputSize:DWord, InputArray_Ptr:DWord, OutputArray:DWord
;---------SET UP FPU FOR OPERATION----------

	Fsave FPUState							;Just save the whole bang shoot
	Fwait									;zzz...
	Finit									;BANG!!
	Fwait									;zzz...

	Fstcw FPUControlReg						;Load Control word
	Fwait									;zzz...
	Mov Eax, CR_PrecisionAND				;Select Real8 precision mask
	And Eax, FPUControlReg					;Clr precision bits
	Or Eax, CR_PrecisionReal8				;Set Real8 precision
	Mov FPUControlReg, Eax					;Put in memory location
	Fldcw FPUControlReg						;Real8 mode
	Fwait									;zzz...

;----LOAD SOURCE AND DESTINATION POINTERS---
	Mov Esi, InputArray_Ptr					;Signal array source
	Sub Esi, 4								;Zero base adjust
	Mov Edi, OutputArray_Ptr				;
	Sub Edi, 8								;Zero base adjust
	Mov Ecx, InputSize						;

Con32Loop:
	Fild DWord Ptr [Esi + Ecx * 4]			;Load 32bit integer
	Fwait									;zzz...
	Fstp QWord Ptr [Edi + Ecx * 8]			;Save as 64bit Real8
	Fwait									;
	Loop Con32Loop							;

;------------RESTORE FPU STATE--------------
	Frstor FPUState							;Restore FPU state
	Fwait									;zzz...

	Mov Eax, TRUE							;
	Ret 									;
Convert_Int32_Real8 EndP
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;Real8_FFT:		Inplace FFT of 2 input arrays
;		   		Calculates Complex DFT
;
;Prerequisites:	1)All values are 64bit (Real8) format
;			   	2)Input arrays are BitReversed
;			   	3)The output array ptr is the inplace destination
;-----------------------------------------------------------------------------
Real8_FFT Proc InputSize:DWord, InputArray_RPtr:DWord, InputArray_IPtr:DWord

;---------SET UP FPU FOR OPERATION----------

	Fsave FPUState							;Just save the whole bang shoot
	Fwait									;zzz...
	Finit									;BANG!!
	Fwait									;zzz...

	Fstcw FPUControlReg						;Load Control word
	Fwait									;zzz...
	Mov Eax, CR_PrecisionAND				;Select Real8 precision mask
	And Eax, FPUControlReg					;Clr precision bits
	Or Eax, CR_PrecisionReal8				;Set Real8 precision
	Mov FPUControlReg, Eax					;Put in memory location
	Fldcw FPUControlReg						;Real8 mode
	Fwait									;zzz...

;----LOAD SOURCE AND DESTINATION POINTERS---
	Mov Esi, InputArray_RPtr				;Input array Real ptr
	Mov Edi, InputArray_IPtr				;Input array Image ptr

;---LOOP CONTROL VALUES---------------------
	Mov Eax, InputSize						;
	Mov FFT_Inputsize, Eax					;Inputsize
	Dec Eax									;
	Mov FFT_ButterFlyLoopLimit, Eax			;InputSize-1

;---STAGE LOOP LIMIT -----------------------
	Fld1									;(0)1.0
	Fld FFT_Inputsize						;(0)Arraysize,(1)1.0
	Fyl2x									;(0)1*Log2(FFT_Inputsize)
	Fstp FFT_StageLoopLimit					;Barrel clear

;*******************************************
;************* STAGELOOPS ******************
	Mov FFT_StageCounter, 1					;StageCounter PreLoad
StageLoop:
	Mov Eax, FFT_StageCounter				;
	Cmp Eax, FFT_StageLoopLimit				;
	Jle @F									;
	Mov Eax, TRUE							;
	Ret 									;FFT Finished

@@:
	Fld1									;(0)1.0
	Fstp FFT_URDbl							;Barrel clear
	Fldz									;(0)0.0
	Fstp FFT_UIDbl							;Barrel clear
;-------------------------------------------
;Butterfly Step = Int(2*EXP(StageLoopCount))
;
;DFTLoopLimit = Butterfly Step \ 2
;-------------------------------------------
	Fld1									;FPU Barrel
	Fld1									;(0)1.0,(2)1.0
	Faddp St(1)								;(0)2.0
	Fild FFT_StageCounter					;(0)StageCount, (1)2.0
	Fldl2e									;(0)Log2(e),(1)FFTSize,(2)2.0
	Fmul St(1)								;(0)FFTSize*Log2(e),(1)FFTSize,(2)=2.0
	Fst St(1)								;(0)FFTSize*Log2(e),(1)FFTSize*Log2(e),(2)2.0
	Frndint									;(0)Int(St(0)),(1)=FFTSize*Log2(e),(2)2.0
	Fxch									;(0)FFTSize*Log2(e),(1)Int(St(0)),(2)2.0
	Fsub St(1)								;(0)Fraction part,(1)Int part,(2)2.0
	F2xm1									;(0)2^Fraction part-1,(1)Int part,(2)2.0
	Fld1									;(0)1.0,(1)2^Fraction part-1,(2)Int part,(3)2.0
	Faddp St(1)								;(0)2^Fraction part,(1)Int part,(2)2.0
	Fscale									;(0)Scaled,(1)Int part,(2)2.0
	Fmulp St(2)								;(0)Int part,(1)Ans*2
	Fstp St(0)								;(0)Ans*2
	Fst FFT_ButterFlyStep					;FFT_ButterFlyStep = Int(2*EXP(StageLoopCount))
	Fstp FFT_DFTLoopLimit					;Barrel clear
	Fwait									;Wag n bietjie
	Shr FFT_DFTLoopLimit, 1					;DFTLoopLimit\2
;-------------------------------------------
;Sin(PI / DFTLoopLim)
;Cos(PI / DFTLoopLim)
;-------------------------------------------
	Fldpi 									;(0)PI
	Fild FFT_DFTLoopLimit					;(0)DFTLoopLim, (1)PI
	Fdivp St(1), St(0)						;(0)PI/DFTLoopLim

	Fsincos									;(0)Cos(PI/DFTLoopLim), (1)Sin(PI/DFTLoopLim)
	Fstp FFT_CosTmp							;(0)Sin(PI/DFTLoopLim)
	Fstp FFT_SinTmp							;Barrel clear
	Fwait									;
;*******************************************
;************* DFT LOOPS *******************
	Mov FFT_DFTLoopCounter, 1				;PreLoad DFTLoopCounter
DFT_Loop:
	Mov Ecx, FFT_DFTLoopCounter				;
	Cmp Ecx, FFT_DFTLoopLimit				;DFTLoopCounter > DFTLoopLimit ?
	Jle @F									;
	Inc FFT_StageCounter					;Next StageCounter
	Jmp StageLoop							;
;*******************************************
;********** BUTTERFLY LOOP *****************
@@:
	Dec Ecx									;Adjust for ButterFlyCounter
   	Mov FFT_ButterFlyCounter, Ecx			;

ButterFly_Loop:
	Cmp Ecx, FFT_ButterFlyLoopLimit			;ButterFlyCounter > ButterFlyLoopLimit ?
	Jle @F									;Nope
	Inc FFT_DFTLoopCounter					;Next DFTLoopCounter
;-------------------------------------------
;UR = UR*CosTmp - UI*SinTmp
;-------------------------------------------
	Fld FFT_URDbl							;(0)FFT_URDbl
	Fld FFT_CosTmp							;(0)FFT_Cos, (1)FFT_URDbl
	Fmulp St(1)								;(0)FFT_Cos*FFT_URDbl

	Fld FFT_UIDbl							;(0)FFT_UIDbl,(1)FFT_Cos*FFT_URDbl
	Fld FFT_SinTmp							;(0)FFT_Sin, (1)FFT_UIDbl, (2)FFT_Cos*FFT_URDbl
	Fmulp St(1)								;(0)FFT_Sin*FFT_UIDbl,(1)FFT_Cos*FFT_URDbl
	Fsubp St(1)								;(0)=1-0
	Fstp FFT_URDbl							;Barrel clear
;-------------------------------------------
;UI = UR*SinTmp + UI*CosTmp
;-------------------------------------------
	Fld FFT_URDbl							;(0)FFT_URDbl
	Fld FFT_SinTmp							;(0)FFT_Sin, (1)FFT_URDbl
	Fmulp St(1)								;(0)FFT_Sin*FFT_URDbl

	Fld FFT_UIDbl							;(0)FFT_UIDbl,(1)FFT_Sin*FFT_URDbl
	Fld FFT_CosTmp							;(0)FFT_Cos, (1)FFT_UIDbl, (2)FFT_Sin*FFT_URDbl
	Fmulp St(1)								;(0)FFT_Cos*FFT_UIDbl,(1)FFT_Sin*FFT_URDbl
	Faddp St(1)								;(0)=1+0
	Fstp FFT_UIDbl							;Barrel clear
	Fwait

	Jmp DFT_Loop							;Next DFT Loop
;----------Butterfly Loop Continue----------
@@:
	Mov Eax, Ecx							;ECX = Butterfly Loop counter
	Add Eax, FFT_DFTLoopLimit				;Eax = IP pointer
;-------------------------------------------
;Esi, InputArray_RealPtr
;Edi, InputArray_ImagePtr
;Eax, IP pointer
;Ecx, I pointer (butterfly loop counter)
;-------------------------------------------
;-------------------------------------------
;TmpDbl1: = Real[eax]*UR - Image[eax]*UI
;-------------------------------------------
	Fld FFT_URDbl							;(0)FFT_URDbl
	Fld QWord Ptr [Esi + Eax * 8]			;(0)RealR8, (1)FFT_URDbl
	Fmulp St(1)								;(0)RealR8*FFT_URDbl

	Fld FFT_UIDbl							;(0)FFT_UIDbl,(1)RealR8*FFT_URDbl
	Fld QWord Ptr [Edi + Eax * 8]			;(0)ImageR8,(1)FFT_UIDbl,(2)RealR8*FFT_URDbl
	Fmulp St(1)								;(0)ImageR8*FFT_UIDbl, ,(1)RealR8*FFT_URDbl

	Fsubp St(1)								;(0)TmpDbl1 = (1) - (0)
;-------------------------------------------
;TmpDbl2: = Real[eax]*UI + Image[eax]*UR
;-------------------------------------------
	Fld FFT_UIDbl							;(0)FFT_UIDbl, (1)TmpDbl1
	Fld QWord Ptr [Esi + Eax * 8]			;(0)RealR8, (1)FFT_UIDbl, (2)TmpDbl1
	Fmulp St(1)								;(0)RealR8*FFT_UIDbl, (1)TmpDbl1

	Fld FFT_URDbl							;(0)FFT_URDbl,(1)RealR8*FFT_UIDbl, (2)TmpDbl1
	Fld QWord Ptr [Edi + Eax * 8]			;(0)ImageR8,(1)FFT_URDbl,(2)RealR8*FFT_UIDbl, (3)TmpDbl1
	Fmulp St(1)								;(0)ImageR8*FFT_URDbl, ,(1)RealR8*FFT_UIDbl, (2)TmpDbl1

	Faddp St(1)								;(0)TmpDbl2, (1)TmpDbl1
;-------------------------------------------
;Real[eax] = Real[ecx] - TmpDbl1
;-------------------------------------------
	Fld QWord Ptr [Esi + Ecx * 8]			;(0)Real8, (1)TmpDbl2, (2)TmpDbl1
	Fld St(2)								;(0)TmpDbl1, (1)Real8, (2)TmpDbl2, (3)TmpDbl1
	Fsubp St(1)								;(0)=1-0, (1)TmpDbl2, (2)TmpDbl1
	Fstp QWord Ptr [Esi + Eax * 8]			;(0)TmpDbl2, (1)TmpDbl1
;-------------------------------------------
;Image[eax] = Image[ecx] - TmpDbl2
;-------------------------------------------
	Fld QWord Ptr [Edi + Ecx * 8]			;(0)Image8, (1)TmpDbl2, (2)TmpDbl1
	Fld St(0)								;(0)TmpDbl2, (1)Image8, (2)TmpDbl2, (3)TmpDbl1
	Fsubp St(1)								;(0)=1-0, (1)TmpDbl2, (2)TmpDbl1
	Fstp QWord Ptr [Edi + Eax * 8]			;(0)TmpDbl2, (1)TmpDbl1
;-------------------------------------------
;Image[ecx] = Image[ecx] + TmpDbl2
;-------------------------------------------
	Fld QWord Ptr [Edi + Ecx * 8]			;(0)Image8, (1)TmpDbl2, (2)TmpDbl1
	Faddp St(1)								;(0)Image8+TmpDbl2, (2)TmpDbl1
	Fstp QWord Ptr [Edi + Ecx * 8]			;(0)TmpDbl1
;-------------------------------------------

;Real[ecx] = Real[ecx] + TmpDbl1
;-------------------------------------------
	Fld QWord Ptr [Esi + Ecx * 8]			;(0)Real8, (1)TmpDbl1
	Faddp St(1)								;(0)Real8+TmpDbl1
	Fstp QWord Ptr [Esi + Ecx * 8]			;Barrel clear
	Fwait									;zzzzzz...

	Add Ecx, FFT_ButterFlyStep				;Adjust ECX

	Jmp ButterFly_Loop
;********** BUTTERFLY LOOP *****************
;*******************************************
Real8_FFT EndP
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------



